from collections import defaultdict

from obspy import Stream
from obspy.signal.filter import envelope
import numpy as np


def piecewise_from_thresholds(x, thresholds):
    if thresholds:
        _, value = thresholds[0]
        for threshold, value in reversed(thresholds):
            if x >= threshold:
                break
        return value
    else:
        raise RuntimeError()


def zip_streams(stream1, stream2):
    common_ids = set.intersection({trace.id for trace in stream1}, {trace.id for trace in stream2})
    return list(zip(Stream([trace for trace in stream1 if trace.id in common_ids]),
                    Stream([trace for trace in stream2 if trace.id in common_ids])))


def relative_pick_time(stats1, stats2, pick1, pick2, shift):
    pick_time1 = pick1 - stats1.starttime
    pick_time2 = pick2 - stats2.starttime + shift * stats2.delta
    return (pick_time1 + pick_time2) / 2


def cc_preprocess(trace, pick_p_mean_delay, pick_s_mean_delay, freq_range, full_waveform_window, taper_length=0.15):
    new_trace = trace.copy()
    starttime = trace.stats.starttime + pick_p_mean_delay - full_waveform_window[0]
    endtime = trace.stats.starttime + pick_s_mean_delay + full_waveform_window[1]
    new_trace.detrend("constant")
    new_trace.filter("bandpass", freqmin=freq_range[0], freqmax=freq_range[1], corners=2, zerophase=True)
    new_trace.taper(taper_length)
    new_trace.trim(starttime=starttime, endtime=endtime)
    return new_trace


def trim_window(trace, center, window):
    trace.trim(starttime=trace.stats.starttime + center - window[0],
               endtime=trace.stats.starttime + center + window[1],
               pad=True,
               fill_value=0.0)


def estimate_s_pick(trace, p_pick, distance, params):
    vp = params['Vp']
    vs = params['Vs']
    window = params['s_pick_estimation_window']
    shift = params['s_pick_estimation_shift']
    freqmin = params['s_pick_estimation_freqmin']
    freqmax = params['s_pick_estimation_freqmax']
    trace_filt = trace.copy()
    trace_filt.filter("bandpass", freqmin=freqmin, freqmax=freqmax, corners=2, zerophase=True)
    data_envelope = envelope(trace_filt.data)
    sample_shift = int(shift * trace.stats.sampling_rate)

    p_pick_sample = int((p_pick - trace.stats.starttime) * trace.stats.sampling_rate)
    delta_ps_thumble_rule_sample = int(distance * (vp - vs) / (vp * vs) * trace.stats.sampling_rate)
    s_pick_thumble_rule_sample = p_pick_sample + delta_ps_thumble_rule_sample

    delta1, delta2 = window
    start = s_pick_thumble_rule_sample - delta1
    stop = s_pick_thumble_rule_sample + delta2
    if start < 0 or stop > data_envelope.size:
        raise ValueError("Envelope out of bound (caused by rule of thumb)")
    s_pick_sample = start + np.argmax(data_envelope[start:stop]) - sample_shift

    return trace.stats.starttime + s_pick_sample * trace.stats.delta


def longest_subsequence_of_trues(bools):
    if len(bools) > 0:
        sequences = [None]
        for n, x in enumerate(bools):
            current_sequence = sequences[-1]
            if x:
                if current_sequence is None:
                    sequences[-1] = (n, n)
                else:
                    start, _ = current_sequence
                    sequences[-1] = (start, min(n + 1, len(bools)))
            else:
                if current_sequence is not None:
                    sequences.append(None)
        if sequences[-1] is None:
            sequences.pop()
        if sequences:
            return max(sequences, key=lambda pair: pair[1] - pair[0])
        else:
            return None
    else:
        return None


def connected_components(edges):
    neighbors = defaultdict(set)
    for edge in edges:
        for node in edge:
            neighbors[node].update(edge)
    graph = set(neighbors.keys())
    visited = set()

    def component_from(starting_node):
        component = {starting_node}
        while component:
            current_node = component.pop()
            visited.add(current_node)
            component |= (neighbors[current_node] - visited)
            yield current_node

    components = []
    for node in graph:
        if node not in visited:
            components.append(sorted(component_from(node)))

    return components



import logging
from itertools import combinations
from math import cos, hypot

import numpy as np
import obspy
import pandas as pd
from obspy import UTCDateTime
from obspy.geodetics import gps2dist_azimuth

from obspy.signal.trigger import pk_baer

from . import custom_formats
from .utils import estimate_s_pick


class InventoryLookupError(BaseException):
    pass


class MissingPhaseDataError(BaseException):
    pass


class SPickEstimationError(BaseException):
    pass


def zmap(catalog_path, extensions=None):
    names = ['longitude',
             'latitude',
             'year',
             'month',
             'day',
             'magnitude',
             'depth',
             'hour',
             'minute',
             'second']

    if extensions:
        names.extend(extensions)

    catalogue = pd.read_csv(catalog_path,
                            sep=r'\s+',
                            usecols=range(len(names)),
                            names=names,
                            parse_dates={'date': ['year', 'month', 'day', 'hour', 'minute', 'second']},
                            date_parser=lambda datestr: pd.to_datetime(datestr, format='%Y %m %d %H %M %S.%f',
                                                                       utc=True))
    return catalogue


# TODO: improve performance, reading of phase_file should be cached (also, only hypoinv has been thoroughly tested)
def errors(catalogue, phase_file, phase_type):
    if phase_type == 'hypoinv':
        error_function = _hypoinverse_errors
    elif phase_type == 'hypoel':
        error_function = _hypoel_errors
    elif phase_type in ('quakeml', 'nll'):
        error_function = _obspy_errors
    elif phase_type in custom_formats.registry:
        error_function = custom_formats.registry[phase_type]['errors']
    else:
        raise NotImplementedError()

    return {event.name: error_function(event.date, phase_file) for event in catalogue.itertuples()}


def _textfile_errors(origin_time, phase_file, columns_delimiters):
    error_values = {}
    with phase_file.open('r') as file:
        while line := file.readline():
            if line.startswith(origin_time.strftime('%Y%m%d%H%M%S')):
                for key, (a, b) in columns_delimiters.items():
                    try:
                        error_values[key] = _read_float(line, a, b)
                    except ValueError as err:
                        logging.warning("Found invalid string.", exc_info=err)
                        error_values[key] = None
    return error_values


# Column numbers taken from the documentation of the hypoinverse/hypoel formats
def _hypoinverse_errors(origin_time, phase_file):
    return _textfile_errors(origin_time, phase_file, {'time_uncertainty': (48, 52),
                                                      'horizontal_uncertainty': (85, 89),
                                                      'vertical_uncertainty': (89, 93)})


def _hypoel_errors(origin_time, phase_file):
    return _textfile_errors(origin_time, phase_file, {'time_uncertainty': (47, 50),
                                                      'horizontal_uncertainty': (56, 59),
                                                      'vertical_uncertainty': (74, 77)})


def _obspy_errors(origin_time, phase_file):
    catalogue = obspy.read_events(str(phase_file))
    error_values = {}
    for event in catalogue:
        if abs(event.time - obspy.UTCDateTime(origin_time)) < 1.0:
            origin, = event.origins
            error_values['time_uncertainty'] = origin.time_errors.uncertainty
            error_values['horizontal_uncertainty'] = _horizontal_error(origin)
            error_values['vertical_uncertainty'] = origin.depth_errors.uncertainty
    return error_values


def _horizontal_error(origin, earth_radius_km=6.371e3):
    lat = origin.latitude
    sigma_lat = origin.latitude_errors.uncertainty
    sigma_lon = origin.longitude_errors.uncertainty
    return (np.pi / 180.0) * earth_radius_km * hypot(sigma_lat, sigma_lon * cos(lat))


def _read_float(whole_str, start, end, decimals=2):
    if whole_str[start:end].isspace():
        return None
    float_str = whole_str[start:end - decimals] + '.' + whole_str[end - decimals:end]
    float_str = float_str.replace(' ', '0')
    return float(float_str)


def coordinates(inventory, network_code, station_code, time, tol=0.2):
    stations = []
    for network in inventory.select(network=network_code, station=station_code, time=time):
        for station in network:
            stations.append(station)
    if stations:
        if len(stations) > 1:
            maxdist = 1e-3 * np.max([gps2dist_azimuth(s.latitude, s.longitude, r.latitude, r.longitude)[0]
                                     for s, r in combinations(stations, 2)])
            if maxdist > tol:
                raise InventoryLookupError(f"Incompatible matches for {station_code} at {time} found in inventory.")
        return stations[0].latitude, stations[0].longitude
    else:
        raise InventoryLookupError(f"Station {station_code} at {time} cannot be found.")


def picks(event, event_coordinates, station_coordinates, trace, params, phase_file=None, phase_type=None, picker=False,
          picker_arguments=None, travel_times_function=None):
    station_latitude, station_longitude = station_coordinates

    if phase_file is None:
        if picker:
            if picker_arguments is None:
                raise MissingPhaseDataError("Picker arguments must be provided")
            p_pick_sample_delay, _ = pk_baer(trace.data, trace.stats.sampling_rate, *picker_arguments)
            p_pick = trace.stats.starttime + trace.stats.delta * p_pick_sample_delay
            s_pick = None
        elif travel_times_function is not None:
            p_pick, s_pick = compute_picks(event, station_latitude, station_longitude, travel_times_function,
                                           p_model_corr=params['p_velocity_model_correction'],
                                           s_model_corr=params['s_velocity_model_correction'])
        else:
            raise MissingPhaseDataError("Either a phase file, picker or a model must be provided")
    else:
        p_pick, s_pick = _read_picks(event.date, trace.stats.station, phase_file, phase_type)
        if p_pick is None:
            if picker:
                if picker_arguments is None:
                    raise MissingPhaseDataError("Picker arguments must be provided")
                p_pick_sample_delay, _ = pk_baer(trace.data, trace.stats.sampling_rate, *picker_arguments)
                p_pick = trace.stats.starttime + trace.stats.delta * p_pick_sample_delay
                s_pick = None
            elif travel_times_function is not None:
                p_pick, s_pick = compute_picks(event, station_latitude, station_longitude, travel_times_function,
                                               p_model_corr=params['p_velocity_model_correction'],
                                               s_model_corr=params['s_velocity_model_correction'])
            else:
                raise MissingPhaseDataError("Phase file with picking for P phase, picker or a model must be provided")
        if s_pick is None:
            event_latitude, event_longitude = event_coordinates
            epi_dist, _, _ = gps2dist_azimuth(event_latitude, event_longitude, station_latitude, station_longitude)
            try:
                s_pick = estimate_s_pick(trace, p_pick, epi_dist, params)
            except ValueError as err:
                raise SPickEstimationError(str(err))
    return p_pick, s_pick


# TODO: improve performance, phase_file should be cached (also, only hypoinv has been thoroughly tested)
def _read_picks(origin_time, station, phase_file, phase_type):
    if phase_type == 'hypoinv':
        phases_function = _hypoinverse_phases
    elif phase_type == 'hypoel':
        phases_function = _hypoel_phases
    elif phase_type in ('nll', 'quakeml'):
        phases_function = _ob_phases
    elif phase_type in custom_formats.registry:
        phases_function = custom_formats.registry[phase_type]['phases']
    else:
        raise NotImplementedError()

    return phases_function(origin_time, station, phase_file)


def _hypoinverse_phases(origin_time, station, phase_file):
    with phase_file.open('r') as file:
        lines = []
        start_section = False
        while line := file.readline():
            if line.startswith(origin_time.strftime('%Y%m%d%H%M%S')):
                start_section = True
                continue
            if start_section and line.startswith(30 * " "):
                break
            if start_section:
                lines.append(line)
    p_pick, s_pick = None, None
    for line in lines:
        if line.startswith(station) and int(line[16]) < 4:
            if not line[30:34].isspace():
                p_pick = UTCDateTime.strptime(line[17:34], '%Y%m%d%H%M %S%f')
            elif not line[42:46].isspace():
                s_pick = UTCDateTime.strptime(line[17:30] + line[42:46], '%Y%m%d%H%M %S%f')
    return p_pick, s_pick


def _hypoel_phases(origin_time, station, phase_file):
    with phase_file.open('r') as file:
        lines = []
        start_section = False
        while line := file.readline():
            if line.startswith(origin_time.strftime('%Y%m%d%H%M%S')):
                start_section = True
                continue
            if start_section and line.startswith(17 * " "):
                break
            if start_section:
                lines.append(line)
    p_pick, s_pick = None, None
    for line in lines:
        if line.startswith(station) and int(line[7]) < 4:
            if not line[20:24].isspace():
                p_pick = UTCDateTime.strptime(line[9:19] + line[20:24], '%y%m%d%H%M%S%f')
            if not line[32:36].isspace():
                s_pick = UTCDateTime.strptime(line[9:19] + line[32:36], '%y%m%d%H%M%S%f')
    return p_pick, s_pick


def _ob_phases(origin_time, station, phase_file):
    catalogue = obspy.read_events(str(phase_file))
    p_pick, s_pick = None, None
    for event in catalogue:
        if abs(event.time - obspy.UTCDateTime(origin_time)) < 1.0:
            for pick in event.picks:
                if pick.waveform_id.station_code == station and pick.phase_hint == 'P' \
                        and pick.time_errors.uncertainty < 1.0:
                    p_pick = pick.time
                if pick.waveform_id.station_code == station and pick.phase_hint == 'S' \
                        and pick.time_errors.uncertainty < 1.0:
                    s_pick = pick.time
    return p_pick, s_pick


def compute_picks(event, station_latitude, station_longitude, model_function, p_model_corr=None, s_model_corr=None,
                  earth_radius=6.371e6, min_event_depth=1.5e3):
    origin_time = UTCDateTime(event.date)
    event_latitude = event.latitude
    event_longitude = event.longitude
    event_depth = event.depth
    event_depth = max(event_depth, min_event_depth)

    epi_dist, _, _ = gps2dist_azimuth(event_latitude, event_longitude, station_latitude, station_longitude)
    deg = (180.0 / np.pi) * epi_dist / earth_radius

    arrivals_p = model_function(source_depth_in_km=event_depth * 1e-3, distance_in_degree=deg, phase_list=("p", "P"))
    arrivals_s = model_function(source_depth_in_km=event_depth * 1e-3, distance_in_degree=deg, phase_list=("s", "S"))
    pick_p = origin_time + arrivals_p[0].time
    pick_s = origin_time + arrivals_s[0].time
    if p_pick_correction := p_model_corr:
        pick_p -= p_pick_correction
    if s_pick_correction := s_model_corr:
        pick_s -= s_pick_correction
    return pick_p, pick_s



from matplotlib import figure
from matplotlib.colors import Normalize
from matplotlib.cm import ScalarMappable
import numpy as np
from obspy.geodetics import gps2dist_azimuth

from .read import coordinates


def signals(trace1, trace2, p_wave_window_1, p_wave_window_2, s_wave_window_1, s_wave_window_2, text, filename):
    fig = figure.Figure()
    ax = fig.add_subplot()
    ax.plot(np.arange(trace1.stats.npts) * trace1.stats.delta, trace1.data / abs(trace1.max()), linewidth=1,
            color="blue", alpha=0.5)
    ax.plot(np.arange(trace2.stats.npts) * trace2.stats.delta, trace2.data / abs(trace2.max()), linewidth=1,
            color="red", alpha=0.5)
    ax.set_xlabel("Times (sec)")
    ax.set_ylabel("Normalized Amplitude")

    ax.text(0.5, 1, text)
    ax.set_ylim(-1.3, 1.3)
    ax.hlines(-1, p_wave_window_1, p_wave_window_2, color="r", linewidth=1, linestyles="dashed")
    ax.hlines(-1, s_wave_window_1, s_wave_window_2, color="r", linewidth=1, linestyles="dashed")
    fig.savefig(filename)


def cross_spectrum(trace1, trace2, cs, slope, indices, textline, filename):
    fig = figure.Figure(constrained_layout=True)
    waveforms, analysis = fig.subfigures(2, 1)

    waveform1, waveform2 = waveforms.subplots(2, 1, sharex=True)
    data1 = trace1.data
    data1 = 2 * (data1 - data1.min()) / (data1.max() - data1.min()) - 1
    waveform1.plot(np.arange(trace1.stats.npts) * trace1.stats.delta, data1, linewidth=1)
    waveform1.set_ylabel("Amplitude")
    data2 = trace2.data
    data2 = 2 * (data2 - data2.min()) / (data2.max() - data2.min()) - 1
    waveform2.plot(np.arange(trace2.stats.npts) * trace2.stats.delta, data2, linewidth=1)
    waveform2.set_ylabel("Amplitude")
    waveform2.set_xlabel("Time [sec]")

    coherence, phase = analysis.subplots(2, 1, sharex=True)
    positive_freq_indices = cs.freq.flatten() > 0.0
    x = cs.freq.flatten()
    y = cs.cohe.flatten()
    z = cs.phase.flatten() * np.pi / 180.0
    start, stop = indices
    coherence.plot(x[positive_freq_indices], y[positive_freq_indices], linewidth=1)
    coherence.plot(x[start:stop], y[start:stop], "rx", label="used data", markersize=4)
    coherence.set_ylabel("Coherency")
    coherence.set_ylim(0, 1.1)
    phase.plot(x[positive_freq_indices], z[positive_freq_indices], linewidth=1)
    phase.plot(x[start:stop], z[start:stop], "rx", label="used data", markersize=4)
    phase.plot(x[start:stop], slope * x[start:stop], "g", label="fitted line")
    phase.set_xlabel("Frequency [Hz]")
    phase.set_ylabel("Phase [rad]")
    phase.set_ylim(-3, 3)
    phase.text(1, 1, textline)
    fig.savefig(filename)


def similarity(ev1, ev2, data, inventory, cc_threshold, delta_threshold, filename):
    event_latitude = (ev1.latitude + ev2.latitude) / 2
    event_longitude = (ev1.longitude + ev2.longitude) / 2
    xs = np.fromiter((x for x, _ in data[(ev1.Index, ev2.Index)].values()), dtype=np.float)
    ys = np.fromiter((y for _, y in data[(ev1.Index, ev2.Index)].values()), dtype=np.float)
    zs = 1e-3 * np.fromiter(map(lambda s: gps2dist_azimuth(event_latitude, event_longitude,
                                                           *coordinates(inventory, s[0], s[1], ev1.date))[0],
                                data[(ev1.Index, ev2.Index)]),
                            dtype=np.float)
    fig = figure.Figure()
    ax = fig.add_subplot()
    ax.scatter(xs, np.absolute(ys), c=zs, cmap="jet")
    ax.set_xlim(cc_threshold, 1.0)
    ax.set_ylim(0.0, delta_threshold)
    ax.set_xlabel("CC Z_component")
    ax.set_ylabel("|ΔS-P|")
    ax.set_title(f"{ev1.name}-{ev2.name}")
    cb = fig.colorbar(ScalarMappable(norm=Normalize(vmin=zs.min(), vmax=zs.max()), cmap="jet"))
    cb.set_label("epi station-event distance (km)")
    ax.text(cc_threshold * 1.01, delta_threshold * 0.95, f"M_{ev1.name}={ev1.magnitude}")
    ax.text(cc_threshold * 1.01, delta_threshold * 0.9, f"M_{ev2.name}={ev2.magnitude}")
    fig.savefig(filename)
